<?php
/**
 * Copyright © 2018 Filoblu S.p.A. All rights reserved.
 */

namespace FiloBlu\MappConnector\Console\Command;

use Exception;
use FiloBlu\MappConnector\Helper\FileLoggerHelper;
use FiloBlu\MappConnector\Model\Config;
use FiloBlu\MappConnector\Model\Jobs;
use FiloBlu\MappConnector\Model\Mapp;
use Magento\Framework\App\Area;
use Magento\Framework\App\State;
use Symfony\Component\Console\Command\Command;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Output\OutputInterface;

/**
 * Class Clean
 * @package FiloBlu\MappConnector\Console\Command
 */
class Clean extends Command
{

    /**
     * @var Jobs
     */
    protected $_jobs;

    /**
     * @var Mapp
     */
    protected $_mapp;

    /**
     * @var FileLoggerHelper
     */
    protected $_loggerHelper;

    /**
     * @var State
     */
    protected $_state;

    /**
     * @var Config
     */
    protected $_config;

    /**
     * Clean constructor.
     * @param Mapp $mapp
     * @param Jobs $jobs
     * @param State $state
     * @param FileLoggerHelper $loggerHelper
     * @param Config $config
     * @param string|null $name
     */
    public function __construct(
        Mapp $mapp,
        Jobs $jobs,
        State $state,
        FileLoggerHelper $loggerHelper,
        Config $config,
        string $name = null
    ) {
        $this->_jobs = $jobs;
        $this->_mapp = $mapp;
        $this->_state = $state;
        $this->_loggerHelper = $loggerHelper;
        $this->_config = $config;
        parent::__construct($name);
    }

    /**
     * {@inheritdoc}
     */
    protected function execute(
        InputInterface $input,
        OutputInterface $output
    ) {
        //Area Code Fix
        try {
            $this->_state->setAreaCode(Area::AREA_GLOBAL);
        } catch (Exception $e) {
            //Area code is already set
        }

        //Clean records by configuration properties
        $result = $this->_jobs->cleanRecords($this->_config);
        $this->_loggerHelper->info($result);
    }

    /**
     * {@inheritdoc}
     */
    protected function configure()
    {
        $this->setName('mapp:clean');
        $this->setDescription("Clean old processed jobs (--DISABLED--)");
        parent::configure();
    }
}
