<?php
/**
 * Copyright © 2018 Filoblu S.p.A. All rights reserved.
 */

namespace FiloBlu\MappConnector\Console\Command;

use Exception;
use FiloBlu\MappConnector\Helper\FileLoggerHelper;
use FiloBlu\MappConnector\Model\Jobs;
use FiloBlu\MappConnector\Model\Mapp;
use Magento\Framework\App\Area;
use Magento\Framework\App\State;
use Symfony\Component\Console\Command\Command;
use Symfony\Component\Console\Input\InputArgument;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Output\OutputInterface;

/**
 * Class RestProcess
 * @package FiloBlu\MappConnector\Console\Command
 */
class RestProcess extends Command
{

    /**
     * @var Jobs
     */
    protected $_jobs;

    /**
     * @var Mapp
     */
    protected $_mapp;

    /**
     * @var FileLoggerHelper
     */
    protected $_loggerHelper;

    /**
     * @var State
     */
    protected $_state;

    /**
     * @var array
     */
    protected $_entities = ['customer', 'subscriber'];

    /**
     * RestProcess constructor.
     * @param Mapp $mapp
     * @param Jobs $jobs
     * @param State $state
     * @param FileLoggerHelper $loggerHelper
     * @param string|null $name
     */
    public function __construct(
        Mapp $mapp,
        Jobs $jobs,
        State $state,
        FileLoggerHelper $loggerHelper,
        string $name = null
    ) {
        $this->_jobs = $jobs;
        $this->_mapp = $mapp;
        $this->_state = $state;
        $this->_loggerHelper = $loggerHelper;
        parent::__construct($name);
    }

    /**
     * {@inheritdoc}
     */
    protected function execute(
        InputInterface $input,
        OutputInterface $output
    ) {
        //Area Code Fix
        try {
            $this->_state->setAreaCode(Area::AREA_GLOBAL);
        } catch (Exception $e) {
            //Area code is already set
        }

        $entity = strtolower($input->getArgument('entity'));

        $result = true;

        $result = $this->processRestJobs($entity);

        if (!$result) {
            $output->writeln("WARNING: The entity specified " . $entity . " doesn't exists");
        }
    }

    public function processRestJobs($entity = null)
    {
        if (!empty($entity)) {
            if (in_array($entity, $this->_entities)) {
                if ($this->_mapp->_config->getConfigValue('mappconnector/general/enable_export_limit_rest') == 1) {
                    $export_limit = $this->_mapp->_config->getConfigValue('mappconnector/general/export_limit_rest');

                    //Get all rest jobs in waiting status
                    $jobs = $this->_jobs
                        ->getCollection()
                        ->addFieldToFilter('status', 'waiting')
                        ->addFieldToFilter('mode', 'rest')
                        ->addFieldToFilter('entity', $entity)
                        ->setPageSize($export_limit)
                        ->setCurPage(1)
                        ->getData();
                } else {
                    //Get all rest jobs in waiting status
                    $jobs = $this->_jobs
                        ->getCollection()
                        ->addFieldToFilter('status', 'waiting')
                        ->addFieldToFilter('mode', 'rest')
                        ->addFieldToFilter('entity', $entity)
                        ->getData();
                }
            } else {
                return false;
            }
        } else {
            if ($this->_mapp->_config->getConfigValue('mappconnector/general/enable_export_limit_rest') == 1) {
                $export_limit = $this->_mapp->_config->getConfigValue('mappconnector/general/export_limit_rest');

                //Get all rest jobs in waiting status
                $jobs = $this->_jobs
                    ->getCollection()
                    ->addFieldToFilter('status', 'waiting')
                    ->addFieldToFilter('mode', 'rest')
                    ->setPageSize($export_limit)
                    ->setCurPage(1)
                    ->getData();
            } else {
                //Get all rest jobs in waiting status
                $jobs = $this->_jobs
                    ->getCollection()
                    ->addFieldToFilter('status', 'waiting')
                    ->addFieldToFilter('mode', 'rest')
                    ->getData();
            }
        }

        if (count($jobs)) {
            foreach ($jobs as $j) {
                $this->_loggerHelper->debug('');
                $this->_loggerHelper->debug('-------- Start to process ' . ucfirst($j['entity']) . ' ' . $j['entity_id'] . ' --------');
                //Default operation type Create or Update
                $this->_mapp->setJob($j);
                //Make request update job and reset log
                $result = $this->_jobs->updateJobAndResetLog($j, $this->_mapp->getRequest(), $this->_mapp->makeRequest());
                $this->_loggerHelper->debug($result);
            }
        } else {
            $this->_loggerHelper->debug('-------- There are no record to process --------');
        }
    }

    /**
     * {@inheritdoc}
     */
    protected function configure()
    {
        $this->setName("mapp:rest:process");
        $this->setDescription("Process waiting Mapp REST jobs");
        $this->addArgument('entity', InputArgument::OPTIONAL, __('Please specify an Entity'));
        parent::configure();
    }
}
