<?php
/**
 * Copyright © 2018 Filoblu S.p.A. All rights reserved.
 */

namespace FiloBlu\MappConnector\Model\From\Entity;

use Exception;
use FiloBlu\MappConnector\Model\Config;
use FiloBlu\MappConnector\Model\To\Mapp\SFTP;
use Magento\Catalog\Api\ProductRepositoryInterface;
use Magento\Eav\Api\AttributeSetRepositoryInterface;
use Magento\Framework\App\ResourceConnection;
use Magento\Framework\Exception\NoSuchEntityException;
use Magento\Framework\UrlInterface;
use Magento\Quote\Api\CartRepositoryInterface;
use Magento\Quote\Model\Quote\Item;

/**
 * Class QuoteItem
 * @package FiloBlu\MappConnector\Model\From\Entity
 */
class QuoteItem extends AbstractEntity implements EntityInterface
{

    //TODO: Get only abandoned cart

    /**
     * Entity
     *
     * @var string
     */
    protected $_entity = 'quote_item';

    /**
     * Magento Entity Model
     *
     * @var string
     */
    protected $_entityModel = Item::class;

    /**
     * Magento Entity Model Id
     *
     * @var string
     */
    protected $_entityModelPrimary = 'quote_id';

    /**
     * Magento Entity Model Related Id
     *
     * @var string
     */
    protected $_entityModelRelatedPrimary = 'quote_id';

    /**
     * Mapp Resource Model
     *
     * @var string
     */
    protected $_entityMappResource = SFTP::class;

    /**
     * Mapp Resource Model Id
     *
     * @var array
     */
    protected $_entityMappPrimary = ['primaryDB' => 'quote_id', 'primaryMapp' => 'email'];

    /**
     * @var string
     */
    protected $_xmlConfigPath = 'mappconnector/quote_mapping/quote_item';

    /**
     * @var Quote
     */
    protected $_quote;

    /**
     * @var Config
     */
    protected $_config;

    /**
     * @var ProductRepositoryInterface
     */
    protected $productRepository;

    /**
     * @var ResourceConnection
     */
    protected $resourceConnection;

    /**
     * @var AttributeSetRepositoryInterface
     */
    protected $attributeSet;
    /**
     * @var CartRepositoryInterface
     */
    private $cartRepository;

    /**
     * @var bool
     */
    private $canProcess = false;

    /**
     * QuoteItem constructor.
     * @param Quote $quote
     * @param Config $config
     * @param ProductRepositoryInterface $productRepository
     * @param ResourceConnection $resourceConnection
     * @param CartRepositoryInterface $cartRepository
     * @param AttributeSetRepositoryInterface $attributeSet
     */
    public function __construct(
        Quote $quote,
        Config $config,
        ProductRepositoryInterface $productRepository,
        ResourceConnection $resourceConnection,
        CartRepositoryInterface $cartRepository,
        AttributeSetRepositoryInterface $attributeSet
    ) {
        parent::__construct($resourceConnection);
        $this->_config = $config;
        $this->_quote = $quote;
        $this->productRepository = $productRepository;
        $this->attributeSet = $attributeSet;
        $this->cartRepository = $cartRepository;
    }

    /**
     * @param $record
     * @return array @inheritDoc
     * @throws NoSuchEntityException
     */
    public function getDefaultData($record)
    {
        $quote = $this->cartRepository->get($record['quote_id']);
        $store = $quote->getStore();
        //Get Product
        $product = $this->productRepository->getById($record['product_id'], false, $record["store_id"]);

        //Get Product Thumbnail
        $thumbnail = ($product->getThumbnail() === 'no_selection') ? '' : $store->getBaseUrl(UrlInterface::URL_TYPE_MEDIA) . 'catalog/product' . $product->getThumbnail();

        //Get Attribute Set Name
        $attributeSetId = $product->getAttributeSetId();
        $attributeSetName = '';
        if (!empty($attributeSetId)) {
            $attributeSet = $this->attributeSet->get($attributeSetId);
            $attributeSetName = $attributeSet->getAttributeSetName();
        }

        // Populate $result with mandatory data

        return [
            'key'                    => strtolower($quote->getCustomer()->getEmail()??''),
            'product_uri'            => $product->getProductUrl(),
            'product_main_thumbnail' => $thumbnail,
            'product_attribute_set'  => $attributeSetName
        ];
    }

    /**
     * @param $primary
     * @param $id
     * @return array @inheritDoc
     * @throws Exception
     */
    public function getObjectData($primary, $id)
    {
        $connection = $this->resourceConnection->getConnection();
        $isActiveQuery  = $connection->select()->from('quote', ['is_active'])->where('entity_id = ?', $id);
        $this->canProcess = (bool)$connection->fetchOne($isActiveQuery);

        if (!$this->canProcess) {
            return [];
        }

        $tableName = $this->resourceConnection->getTableName('quote_item');
        $only_simple = $this->_config->getConfigValue('mappconnector/quote_mapping/export_only_simple');
        $sql = "SELECT * FROM {$tableName} WHERE {$primary} = {$id}";

        if ($only_simple == 1) {
            $sql .= ' AND product_type = "simple" ';
        }

        return $connection->fetchAll($sql);
    }

    /**
     * @param $record
     * @return array
     * @throws Exception
     */
    public function getAdditionalData($record)
    {
        $map = $this->_config->getNormalizedConfigObjectByPath('mappconnector/quote_mapping/quote_item_product', 'quote_product');

        $result = $this->getRelatedModelData($this->_quote, $record['quote_id']);

        $p = [];

        try {
            $product = $this->productRepository->getById($record['product_id'], false, $record['store_id']);

            foreach ($map as $k => $m) {
                $v = $product->getResource()->getAttribute($k)->getFrontend()->getValue($product);
                if (is_array($v)) {
                    $v = implode(',', $v);
                }
                $p[$m] = (!is_object($v)) ? $v : '';
            }
        } catch (NoSuchEntityException $e) {
            // TODO : log
        }

        return array_merge($result, $p);
    }

    /**
     * @param string $type
     * @param string $last_index
     * @return mixed
     * @throws Exception
     */
    public function getAllData($type = 'full', $last_index = '')
    {
        $connection = $this->resourceConnection->getConnection();
        $tableName = $this->resourceConnection->getTableName('quote'); //gives table name with prefix
        $tableNameJobs = $this->resourceConnection->getTableName('filoblu_mappconnector_jobs');

        if ($type == 'full') {
            //Full
            //$sql = "Select * FROM " . $tableName;
            $sql = "SELECT count(*)
                    FROM
                    {$tableName}
                    WHERE
                    is_active = 1
                    AND
                    items_count > 0
                    AND
                    customer_email IS NOT NULL
                    AND
                    TIMESTAMPDIFF(SECOND,updated_at,NOW()) > " . $this->_config->getConfigValue('mappconnector/quote_mapping/abandoned_cart_time');

            return $connection->fetchAll($sql);
        }

        // TIME_TO_SEC(TIMEDIFF(CURRENT_TIMESTAMP,`down_at` )) > :offset
        //Incremental
        $sql = "SELECT `ce`.`entity_id` as `entity_id`
                    FROM {$tableName} AS `ce`
                    LEFT JOIN {$tableNameJobs} AS `mj`
                    ON
                        `mj`.`entity_id` = `ce`.`entity_id`
                    AND
                        `mj`.`entity` = 'quote'
                    WHERE
                        `mj`.`status` IS NULL
                    AND
                        `ce`.`is_active` = 1
                    AND
                        ce.items_count > 0
                    AND
                        ce.customer_email IS NOT NULL
                    AND
                        TIMESTAMPDIFF(SECOND,ce.updated_at,NOW()) > " . $this->_config->getConfigValue('mappconnector/quote_mapping/abandoned_cart_time');

        return $connection->fetchAll($sql);
    }

    /**
     * @return string
     */
    public function getNormalizedEntityModelPrimary()
    {
        return $this->_quote->getEntityModelPrimary();
    }

    /**
     * @return bool
     */
    public function canProcess()
    {
        return $this->canProcess;
    }
}
