<?php
/**
 * Copyright © 2018 Filoblu S.p.A. All rights reserved.
 */

namespace FiloBlu\MappConnector\Model;

use Exception;
use FiloBlu\MappConnector\Helper\ExceptionHelper;
use FiloBlu\MappConnector\Helper\FileLoggerHelper;
use FiloBlu\MappConnector\Helper\LogHelper;
use Magento\Framework\App\Filesystem\DirectoryList;
use Magento\Framework\App\ObjectManager;
use Magento\Framework\App\ResourceConnection;
use Magento\Framework\Filesystem;
use Magento\Framework\ObjectManagerInterface;

/**
 * Class Mapp
 * @package FiloBlu\MappConnector\Model
 */
class MappCSV
{

    /**
     * Config Model
     *
     * @var Config $config
     */
    protected $_config;

    /**
     * Request object
     *
     * @var object
     */
    protected $_request;

    /**
     * Resource Type
     *
     * @var
     */
    protected $_resource;

    /**
     * Model loaded
     *
     * @var
     */
    protected $_model;

    /**
     * Entity
     *
     * @var
     */
    protected $_entity;

    /**
     * Jobs Model
     *
     * @var
     */
    protected $_jobs;

    /**
     * Job Type
     *
     * @var
     */
    protected $_jobType;

    /**
     * Queried Jobs
     *
     * @var
     */
    protected $_records;

    /**
     * Operation type
     *
     * @var string
     */
    protected $_type = 'createOrUpdate';

    /**
     * @var
     */
    protected $_fileLogger;

    /**
     * @var
     */
    protected $_filename;

    /**
     * @var
     */
    protected $_ids;

    /**
     * @var
     */
    protected $_skippedIds;

    /**
     * @var
     */
    protected $_db_resource;

    /**
     * @var
     */
    protected $_db_connection;

    /**
     * @var
     */
    protected $_db_table_name;
    /**
     * @var string
     */
    protected $_exportPath;
    /**
     * @var Filesystem
     */
    protected $_filesystem;

    /** @var string */
    protected $_primary;
    /**
     * @var false
     */
    protected $file_write;
    /**
     * @var ObjectManagerInterface
     */
    private $objectManager;
    /**
     * @var ResourceConnection
     */
    private $resourceConnection;

    /**
     * Mapp constructor.
     * @param Config $config
     * @param FileLoggerHelper $fileLogger
     * @param Jobs $jobs
     * @param Filesystem $filesystem
     * @param ObjectManagerInterface $objectManager
     * @param ResourceConnection $resourceConnection
     */
    public function __construct(
        Config $config,
        FileLoggerHelper $fileLogger,
        Jobs $jobs,
        Filesystem $filesystem,
        ObjectManagerInterface $objectManager,
        ResourceConnection $resourceConnection
    ) {
        $this->_request = (object)[];
        $this->_fileLogger = $fileLogger;
        $this->_config = $config;
        $this->_jobs = $jobs;
        $this->_filesystem = $filesystem;
        $this->objectManager = $objectManager;
        $this->resourceConnection = $resourceConnection;
    }

    /**
     * Prepare Object and Make the request to the resource
     *
     * @return mixed
     * @throws Exception
     */
    public function makeCSVAndSend()
    {
        $this->_fileLogger->debug('');
        $this->_fileLogger->debug('------------------ Start Process ' . ucfirst($this->_getNormalizedJobType()) . ' -----------------');
        $this->_fileLogger->debug('Make CSV and Send', [], __METHOD__ . ':' . __LINE__);
        if ($this->_config->getConfigValue('mappconnector/export/ftp_export')) {
            //Init object
            $this->_init();
            //Prepare CSV File
            $csvPrepared = $this->_prepareCSV();

            //Get Resource (SFTP)
            $resource = $this->_getMappResource();
            if (!empty($resource)) {
                return $resource->doRequest($this->_request);
            } else {
                $this->_fileLogger->error('No resource loaded', [], __METHOD__ . ':' . __LINE__);
                return null;
            }
        } else {
            echo "--------------------------------------------------------------------------------\n";
            echo "-- The SFTP export is disabled in your config. Please enable it to process records --\n";
            echo "--------------------------------------------------------------------------------\n";
            die();
        }
    }

    /**
     * @return mixed
     */
    private function _getNormalizedJobType()
    {
        $j = explode('_', $this->_jobType);
        return $j[0];
    }

    /**
     * Init Request Object
     */
    protected function _init()
    {
        $this->_fileLogger->debug('Init MappCSV Class', [], __METHOD__ . ':' . __LINE__);
        $this->_config->setEntity($this->_jobType);
        $this->_prepareRequestConfig();
        $this->_initDB();
    }

    /**
     * Prepare config in the request object
     */
    protected function _prepareRequestConfig()
    {
        $this->_fileLogger->debug('Prepare request config', [], __METHOD__ . ':' . __LINE__);
        try {
            $this->_entity = $this->_getEntity();
            $this->_request->config = (object)[];
            $this->_request->config->map = $this->_config->getNormalizedConfigObject();
            $this->_request->config->model = $this->_entity->getEntityModel();
            $this->_request->config->model_primary = $this->_entity->getEntityModelPrimary();
            $this->_request->config->adapter = $this->_config->getAdapter('sftp');
            $this->_request->config->type = $this->_jobType;
            $this->_request->config->normalized_type = $this->_getNormalizedJobType();
            $this->_request->config->mapp_model = '\FiloBlu\MappConnector\Model\To\Mapp\SFTP';
            $this->_request->config->mapp_primary = $this->_entity->getEntityMappPrimary();
            //$this->_model = $entity->getObjectData($this->_request->config->model_primary, $this->_job['entity_id']);
        } catch (Exception $e) {
            $this->_fileLogger->error($e->getMessage(), [], __METHOD__ . ':' . __LINE__);
        }
    }

    /**
     * Get Entity By Type
     *
     * @return mixed
     * @throws Exception
     */
    protected function _getEntity()
    {
        $class = '\FiloBlu\MappConnector\Model\From\Entity\\' . str_replace(' ', '', ucwords(str_replace('_', ' ', $this->_jobType)));
        $this->_fileLogger->debug('Get Entity ' . $class, [], __METHOD__ . ':' . __LINE__);

        if (class_exists($class)) {
            return $this->objectManager->create($class);
        } else {
            $this->_fileLogger->error("The Entity class specified " . $class . " doesn't exists", [], __METHOD__ . ':' . __LINE__);
            return null;
        }
    }

    /**
     * Init DB
     */
    protected function _initDB()
    {
        $this->_db_resource = $this->resourceConnection;
        $this->_db_connection = $this->_db_resource->getConnection();
        $this->_db_table_name = $this->_db_resource->getTableName('filoblu_mappconnector_jobs'); //gives table name with prefix
    }

    /**
     * Prepare CSV File
     *
     * @return bool
     * @throws Exception
     */
    protected function _prepareCSV()
    {

        //Prepare records to process
        $this->_getRecords();

        //Create filename
        $prefix = $this->_config->getConfigValue('mappconnector/general/export_prefix') ?? '';
        $this->_filename = $prefix . $this->_getNormalizedJobType() . '-' . date('Ymd-His') . '.csv';

        //Get export path
        $this->_exportPath = $this->_getExportPath();

        //If path doesn't exists creates it
        if (!file_exists($this->_exportPath)) {
            if (!mkdir($this->_exportPath, 0777, true)) {
                $this->_fileLogger->error('There was an error while creating the folder ' . $this->_exportPath, [], __METHOD__ . ':' . __LINE__);
            }
        }

        //Get records
        $records = $this->_records->getData();

        //If there are records to process
        if (count($records)) {
            $this->_fileLogger->debug('Preparing CSV for ' . $this->_jobType, [], __METHOD__ . ':' . __LINE__);

            //Load Resource
            $resource = $this->_getMappResource();

            //Prevents on line file
            $this->file_write = false;

            foreach ($records as $record) {

                $m = $this->_entity->getObjectData($this->_request->config->model_primary, $record['entity_id']);

                if (!$this->_entity->canProcess()) {
                    $this->_changeStatusAndWriteLog($record, 'skipped', true);
                    continue;
                }

                //If exists at least one record
                if (count($m) > 0) {
                    if ($this->_jobType == 'order_item' || $this->_jobType == 'quote_item') {
                        foreach ($m as $item) {
                            $this->_processRow($resource, $item, $record);
                        }
                    } else {
                        $this->_processRow($resource, $m, $record);
                    }

                    continue;
                }

                $this->_fileLogger->info('Not existing ' . $this->_getNormalizedJobType() . ' record', [], __METHOD__ . ':' . __LINE__);
                $this->_changeStatusAndWriteLog($record, 'skipped', true);

            }

            //Save csv in path
            try {
                if (!$this->file_write) {
                    if (file_exists($this->_exportPath . $this->_filename)) {
                        unlink($this->_exportPath . $this->_filename);
                        return false;
                    }
                }
                return true;
            } catch (Exception $e) {
                $this->_fileLogger->error($e->getMessage() . '. Unable to load file ' . $this->_exportPath . $this->_filename, [], __METHOD__ . ':' . __LINE__);
                return false;
            }
        } else {
            $this->_fileLogger->debug('No record to process for ' . $this->_getNormalizedJobType(), [], __METHOD__ . ':' . __LINE__);
            return false;
        }
    }

    /**
     * Get Records to process
     */
    protected function _getRecords()
    {
        $size = $this->_jobs
            ->getCollection()
            ->addFieldToFilter('status', 'waiting')
            ->addFieldToFilter('mode', 'sftp')
            ->addFieldToFilter('entity', $this->_getNormalizedJobType())
            ->getSize();

        $this->_fileLogger->debug('There are ' . $size . ' ' . $this->_getNormalizedJobType() . ' to process ', [], __METHOD__ . ':' . __LINE__);

        if ($this->_config->getConfigValue('mappconnector/general/enable_export_limit_ftp') == 1) {
            $export_limit = $this->_config->getConfigValue('mappconnector/general/export_limit_ftp');

            $this->_records = $this->_jobs
                ->getCollection()
                ->addFieldToFilter('status', 'waiting')
                ->addFieldToFilter('mode', 'sftp')
                ->addFieldToFilter('entity', $this->_getNormalizedJobType())
                ->setPageSize($export_limit)
                ->setCurPage(1);

            $nexp = $export_limit;
            if ($size < $export_limit) {
                $nexp = $size;
            }

            $this->_fileLogger->debug('Processing ' . $nexp . ' elements ', [], __METHOD__ . ':' . __LINE__);
        } else {
            $this->_records = $this->_jobs
                ->getCollection()
                ->addFieldToFilter('status', 'waiting')
                ->addFieldToFilter('mode', 'sftp')
                ->addFieldToFilter('entity', $this->_getNormalizedJobType());
        }
    }

    /**
     * Get Export Path
     *
     * @return string
     */
    protected function _getExportPath()
    {
        $vardir = $this->_filesystem
            ->getDirectoryRead(DirectoryList::VAR_DIR)
            ->getAbsolutePath();

        return $vardir . $this->_config->getLocalRootDir() . DIRECTORY_SEPARATOR . 'mapp-connector' . DIRECTORY_SEPARATOR . $this->_getNormalizedJobType() . DIRECTORY_SEPARATOR;
    }

    /**
     * Get Mapp Resource By Name
     *
     * @return mixed
     * @throws Exception
     */
    protected function _getMappResource()
    {
        try {
            $class = $this->_request->config->mapp_model;
            $this->_fileLogger->debug('Get MappResource ' . $class, [], __METHOD__ . ':' . __LINE__);
            if (class_exists($class) && !empty($class)) {
                return $this->objectManager->create($class)->getResource();
            } else {
                $this->_fileLogger->error("The Mapp Resource class specified " . $class . " doesn't exists", [], __METHOD__ . ':' . __LINE__);
                return null;
            }
        } catch (Exception $e) {
            $this->_fileLogger->error($e->getMessage(), [], __METHOD__ . ':' . __LINE__);
        }
    }

    /**
     * @param $record
     * @param $status
     * @param bool $write_log
     */
    protected function _changeStatusAndWriteLog($record, $status, $write_log = false)
    {
        $sql_log = '';
        if ($write_log) {
            $log = [];
            if (count(LogHelper::getLogBuffer())) {
                $log = array_merge(['LOGS:', '----------------------------', ''], LogHelper::getLogBuffer(), ['']);
            }
            if (count(ExceptionHelper::getErrorBuffer())) {
                $log = array_merge($log, ['ERRORS:', '----------------------------', ''], ExceptionHelper::getErrorBuffer());
            }
            $logtxt = (count($log)) ? implode('<br/>', $log) : null;

            $status = (ExceptionHelper::getErrorStatus()) ? 'error' : $status;

            ExceptionHelper::resetErrorBuffer();
            ExceptionHelper::resetErrorStatus();
            LogHelper::resetLogBuffer();
            $sql_log = ", log = '".addslashes($logtxt)."'";
        }

        $sql = "UPDATE $this->_db_table_name SET status = '$status' $sql_log WHERE id = " . $record['id'];
        $this->_db_connection->query($sql);
    }

    /**
     * Process CSV Row and Write in file
     *
     * @param $resource
     * @param $m
     * @param $record
     * @return bool
     * @throws Exception
     */
    protected function _processRow($resource, $m, $record)
    {
        if (!in_array($this->_jobType, ['order', 'order_item', 'quote', 'quote_item', 'product'])) {
            if ($resource->_domainIsExcluded($m[$this->_request->config->mapp_primary['primaryDB']])) {
                //Change status in skipped
                $this->_changeStatusAndWriteLog($record, 'skipped', true);
                return false;
            }
        }

        if (count($m) > 0) {
            try {
                //Get default model data
                $default_data = $this->_entity->getDefaultData($m);
                //Get additional from related modules data
                $additional_data = $this->_entity->getAdditionalData($m);

                //Check status of order
                if (in_array($this->_jobType, ['order', 'order_item'])) {
                    $order_statuses = explode(',', $this->_config->getConfigValue('mappconnector/order_mapping/order_statuses_to_export'));
                    if (!in_array($default_data['order_status'], $order_statuses)) {
                        $this->_fileLogger->debug('Skipped by config the status ' . $default_data['order_status'] . ' is not present in ' . implode(',', $order_statuses));
                        //Change status in skipped
                        $this->_changeStatusAndWriteLog($record, 'skipped', true);
                        return false;
                    }
                }

                //Get full record with default, mapped and additional data
                $csvRecord = $this->_config->getMappedArray($this->_request->config->map, array_merge($default_data, $m, $additional_data));

                //Not filtered additional data
                if (!in_array($this->_jobType, ['subscriber'])) {
                    $csvRecord = array_merge($csvRecord, $additional_data);
                }

                //Change status in processed
                $this->_changeStatusAndWriteLog($record, 'processed', true);
                //If is the first record to write
                if (!$this->file_write) {
                    //Write header titles in csv
                    $this->_writeRowInCsv(array_keys($csvRecord));
                }
                //Write row in csv
                $this->_writeRowInCsv($csvRecord);
                //File can be wrote
                $this->file_write = true;
                return true;
            } catch (Exception $e) {
                $this->_fileLogger->error($e->getMessage(), [], __METHOD__ . ':' . __LINE__);
                $this->_changeStatusAndWriteLog($record, 'skipped', true);
                return false;
            }
        } else {
            $this->_fileLogger->info('Not existing ' . $this->_getNormalizedJobType() . ' record', [], __METHOD__ . ':' . __LINE__);
            $this->_changeStatusAndWriteLog($record, 'skipped', true);
            return false;
        }
    }

    /**
     * @param $record
     * @throws Exception
     */
    protected function _writeRowInCsv($record)
    {
        $csvfile = fopen($this->_exportPath . $this->_filename, 'a');
        fwrite($csvfile, implode($this->_config->getConfigValue('mappconnector/general/csv_separator'), $record) . chr(13) . chr(10));
        fclose($csvfile);
    }

    /**
     * Prepare incremental export of Entity
     * @param string $type
     * @param string $mode
     * @throws Exception
     */
    public function prepareExport($type = 'incremental', $mode = 'sftp')
    {
        $this->_fileLogger->debug('Prepare ' . $type . ' export for ' . $this->_getNormalizedJobType(), [], __METHOD__ . ':' . __LINE__);
        $this->_entity = $this->_getEntity();

        //Full incremental let's a full export in a more fast way
        $t = $type;
        if ($type == 'fullincremental') {
            $t = 'incremental';
        }

        $count = $this->_jobs->prepareExport(
            $this->_getNormalizedJobType(),
            $this->_entity->getAllData($t),
            $this->_entity->getNormalizedEntityModelPrimary(),
            $type,
            $mode
        );
        $this->_fileLogger->debug('Has been prepared ' . $count . ' records for ' . $this->_getNormalizedJobType(), [], __METHOD__ . ':' . __LINE__);
        $this->_fileLogger->debug(ucfirst($type) . ' export for ' . $this->_getNormalizedJobType() . ' complete', [], __METHOD__ . ':' . __LINE__);
    }

    /**
     * Set Job in the object
     *
     * @param $jobType
     */
    public function setJobType($jobType)
    {
        $jobt = $jobType;
        if ($jobType == 'order' || $jobType == 'quote') {
            $jobt = $jobType . '_item';
        }
        $this->_jobType = $jobt;
    }

    /**
     * Set Primary in the object
     *
     * @param $primary
     * @throws Exception
     */
    public function setPrimary($primary)
    {
        if (in_array(strtolower(array_keys($primary)), ['primaryDB', 'primaryMapp'])) {
            $this->_primary = $primary;
        } else {
            $this->_fileLogger->error('Wrong associative array. Keys required ["primaryDB","primaryMapp"]', [], __METHOD__ . ':' . __LINE__);
        }
    }

    /**
     * Get the request object
     *
     * @return object
     */
    public function getRequest()
    {
        return $this->_request;
    }

    /**
     * Change selected collection status
     *
     * @param $status
     */
    protected function _changeAllStatus($status)
    {
        $ids = implode(',', $this->_getAllIds());
        $sql = "UPDATE $this->_db_table_name SET status = '$status' WHERE id IN ($ids)";
        $this->_db_connection->query($sql);
    }

    /**
     * GetAll IDS of the Jobs Collection
     *
     * @return array
     */
    protected function _getAllIds()
    {
        $this->_ids = [];
        foreach ($this->_records->getData() as $item) {
            $this->_ids[] = $item['id'];
        }
        return $this->_ids;
    }
}
